/*
 .*.*.*.*.
 .*.*.*.*. Genome scanning using QTLs 
 .*.*.*.*. Jing Hua Zhao
 .*.*.*.*.
 .*.*.*.*. 26-8-1999 drafting
 .*.*.*.*. 31-8-1999 multistage genome scans
 .*.*.*.*.
 */
%macro QTLSIM(H0=0, mapscale=1,h=0.4,p0=0.05,r=0.9,seed=123,rs=0.3,ssize=10000,out=qtlsim);

proc iml;

h0=&h0;                              /*hypothesis indicator*/
mapscale=&mapscale;                  /*magnifying factor*/
h=&h;                                /*total heritability*/
p0=&p0;                              /*the biggest QTL*/
r=&r;                                /*degrading ratio between QTLs*/
rs=&rs;                              /*sib correlation*/
seed=&seed+&i;                       /*random number seed*/
ssize=&ssize;                        /*sample size*/
d=1-(1-r)*h/p0;                      /*denominator as descriminant*/
/*
 * total # of QTLs involved
 */
if d>0 then tqtl=ceil(log(d)/log(r)); 
else do;
   print 'Value of the descriminant =' d;
   print "Invalid h,p,r !!! please reset them";
   abort;
end;

* print 'null/alternative [0/1]' H0,
'mapscale' mapscale, 'heritability' h, 'biggest QTL' p0, 'degrading factor' r,
 'random number seed' seed, 'sib correlation' rs, 'sample size' ssize, 'output to out';

/*
  use only for historical reasons
 */
%*obsolete;

/*
 * Haldane, Kosambi, and Binomial map funtions.
 * Zhao, JH 17/9/97 (IOP)
 */

start haldane(r);
      return (-0.5*log(1-2*r));
finish;
start hinv(x);
      return (0.5*(1-exp(-2*abs(x))));
finish;

/*
 * chromosome lengths, modify to give finer lengths
 * uses Morton chromosome parameters
 */
morton=
{
 294,266,213,238,211,210,178,172,147,174,150,
 179,130,118,151,151,202,143,147,151,92,81,220
};

browman=
{
 290,269,228,212,198,193,182,168,169,173,148,
 171,115,138,122,134,126,126,105,101,58,62,184
};
dib=
{
 292.7,277.0,233.0,212.0,197.6,201.1,184.0,166.4,166.5,181.7,156.1,
 169.1,117.5,128.6,110.2,130.8,128.7,123.8,109.9,96.5,59.6,58.1,198.1
};

chrlen=j(1,23,0);
do i=1 to 23;
   chrlen[i]=ceil(morton[i]/mapscale);
end;
chrpos=j(tqtl,23,0);                 /*chromosome specific positions of QTLs*/
pos=j(2,tqtl,0);                     /*ordered positions [2,] on chromosome[1,]*/
pctqtl=j(1,tqtl,0);                  /*the actual percentage for all QTLs*/
nqtl=j(1,23,0);                      /*# of QTL per chromosome*/
VA=j(tqtl,23,0);                     /*the additive variance of QTLs*/

lgenome=chrlen[+];                   /*genome length using Morgon's*/
longest=chrlen[1];                   /*is chromosome 1*/
beta=j(longest,23,0);                /*betas from Haseman-Elston regression*/
lod=beta;                            /*lod score based on beta*/
sys=beta;                            /*the systematic component*/
VR=beta;                             /*variance of random component*/
clen=j(1,23,0);                      /*cumulative lengths*/
c=0;
do i=1 to 23;
  c=c+chrlen[i];
  clen[i]=c;
end;
pchr=clen/lgenome;
* print 'cumulative length' clen [format=5.0] / 'probability' pchr [format=5.3];
/*
 * step 1: get nqtl[] and chrpos[] on each chromosome
 */
pn=p0;
do i=1 to tqtl;
   j=1;
   x=ranuni(seed);
   do k=1 to 23;
      if x>=clen[k]/lgenome then j=j+1;         /* which chromosome ?     */
   end;
   nqtl[j]=nqtl[j]+1;                          /* increase QTLs there     */
   if j=1 then c=1+(chrlen[j]-1)*x;
   else c=1+(lgenome-1)*x-clen[j-1];
   c=ceil(c);
   * print x j c;
   chrpos[nqtl[j],j]=c;                        /* the nqtl[j]-th QTL on j */
   pos[1,i]=j;
   pos[2,i]=c;
   VA[nqtl[j],j]=pn;
   pn=pn*r;
end;
* print 'position of QTL across chromosomes' / chrpos [format=6.1];

c0=8*(1+rs**2)/ssize;                          /* coefficient of residual   */
do i=1 to 23;                                  /* iterate over chromosomes  */
   points=chrlen[i];
/*
 * step 2: form systematic component
 */
   do k=1 to nqtl[i];                          /* iterate over QTLs, >=0    */
      do j=1 to points;                        /* iterate over points there */
         d=abs(j-chrpos[k,i])/100*mapscale;    /* map distance in Morgans   */
         theta=hinv(d);                        /* recombination probability */
/* 
 * build up the QTL effects 
 */
         if h0=0 then sys[j,i]=0;
         else sys[j,i]=sys[j,i]+(1-2*theta)**2*VA[k,i];
      end;
   end;
/*
 * step 3: variance of the random component for all positions
 */
   do j=1 to points;
      VR[j,i]=c0-sys[j,i]**2/ssize;
   end;
/*
 * steps 4,5: generate normal(0,1) deviate for each positions and random component
 */
  ran=sqrt(VR[1,i])*rannor(seed);
  beta[1,i]=sys[1,i]+ran;
/*
 * step 6: other random points
 */
  do j=2 to points;
     vi=VR[j,i];
     vi1=VR[j-1,i];
     d=1/100*mapscale;                         /*Morgans, change if appropriate*/ 
     theta=hinv(d);                            /*recombination between pos. i,i-1*/
     cc=(1-2*theta)**2;
     c2=cc*sqrt(vi/vi1);                       /*autoregressive coefficient*/
     d2=vi-cc*sqrt(vi*vi1);                    /*random disturbances*/
     ran=sqrt(c2)*ran+sqrt(d2)*rannor(seed);   /*iteration*/
/*
 * tally components of BETA
 */
     beta[j,i]=sys[j,i]+ran;
  end;
/*
 * lod score assumimg no QTL effect
 */
  do j=1 to points;
     if beta[j,i]>0 then lod[j,i]=beta[j,i]**2/2/log(10)/c0;
  end;
end;

maxlod=lod[<>];

/*
 * percentage of lod score, paradoxically lgenome won't do
 * old code: plod=100*(1-rank(lod)/longest/23);
 * or rank(-lod) would do
 * current code is due to 27/8/1999, need to check for overflow
 */
allpoint=j(lgenome,1,0);
k=1;
do i=1 to 23;
   do j=1 to chrlen[i];
      allpoint[k]=lod[j,i];
      k=k+1;
   end;
end;
plod=j(longest,23,0);
k=1;
rankall=rank(-allpoint);
do i=1 to 23;
   do j=1 to chrlen[i];
      plod[j,i]=rankall[k]/lgenome*100;
      k=k+1;
   end;
end;

do i=1 to tqtl;
   pctqtl[i]=plod[pos[2,i],pos[1,i]];
end;

/*
 * present the result below, but now only the percentage
 */

* print beta [format=7.4], lod [format=7.4], seed maxlod pos;

/*
* t=nrow(beta);
* tokeep=j(t,1,h)||j(t,1,p0)||j(t,1,r)||j(t,1,seed)||j(t,1,rs)||j(t,1,ssize)||beta;
* create &out from tokeep;
* append from tokeep;
* close &out;
*/
if h0=0 then do;
   create single from maxlod;
   append from maxlod;
end;
else do;
   create single from pctqtl;
   append from pctqtl;
end;
close single;

quit;
%mend QTLSIM;

%macro obsolete;
start regress;  /*regression for validity checking of correlations*/
  n=nrow(x);                  /* number of observations   */
  k=ncol(x);                  /* number of variables      */
  xpx=x`*x;                   /* cross-products           */
  xpy=x`*y;
  xpxi=inv(xpx);              /* inverse crossproducts    */
  beta=xpxi*xpy;              /* parameter estimates      */
  sse = y`*y-xpy`*beta;       /* sum of squares error     */
  dfe = n-k;                  /* degrees of freedom error */
  mse = sse/dfe;              /* mean square error      */
  rmse = sqrt(mse);           /* root mean square error */
  rsquare = 1-sse/((y-y[:])[##]);
  print ,,'Regression Analysis',,'Residual Error:'
      sse dfe mse rmse rsquare;

  stderr = sqrt(vecdiag(xpxi)#mse); /* std error of estimates  */
  tratio = beta/stderr;             /* test for parameter=0    */
  probt=1-probf(tratio##2,1,dfe);   /* significance probability */
  print ,,'Regression Parameter Estimates ',,
           names beta stderr tratio probt;

  covb=xpxi#mse;              /* covariance of estimates  */
  s=1/stderr;
  corrb=s#covb#s`;            /* correlation of estimates */
  print ,"Covariance of estimates",  covb[r=names c=names],
         "Correlation of estimates",corrb[r=names c=names];
finish;

start corrb(x,y); /*calculation the correlations*/
  n=nrow(x);
  z=j(n,1,0);
  xmean=x[+]/n;
  ymean=y[+]/n;
  sx=0;
  sy=0;
  s=0;
  do i=1 to n;
     sx=sx+(x[i]-xmean)**2;
     sy=sy+(y[i]-ymean)**2;
     s=s+(x[i]-xmean)*(y[i]-ymean);
  end;
  return (s/sqrt(sx)/sqrt(sy));
finish;

start getrho; /*repeat 100 times to get empirical correlation between betas*/

b1=j(100,1,0);
b2=j(100,1,0);
x1=j(1000,1,0);
x2=j(1000,1,0);
y =j(1000,1,0);

do k=1 to 100;
   /*BVN(1/8,1/8,rho) simulations to get rho*/

   mu1=0; mu2=0; var1=1/8; var2=1/8; rho=.5;
   std1=sqrt(var1);  std2=sqrt(var2);

   c=sqrt(1-rho**2);
   do i = 1 to 1000;
      x1[i] = rannor(&seed);
      x2[i] = rho*x1[i]+c*rannor(&seed);
      x1[i] = mu1 + sqrt(var1)*x1[i];
      x2[i] = mu2 + sqrt(var2)*x2[i];
      y[i]=rannor(&seed)*(1-c**2)**2;
   end;

   run regress(y,x1);b1[k]=beta[2];
   run regress(y,x2);b2[k]=beta[2];

end;
rho=corrb(b1,b2);
finish;

start kosambi(r);
      return (0.25*log((1+2*r)/(1-2*r)));
finish;
start kinv(x);
      return (0.5*(exp(4*x)-1)/(exp(4*x)+1));
finish;
start binom(n,r);
      return (0.5*n*(1-(1-2*r)**(1/n)));
finish;
start binv(n,x);
      return (0.5*(1-(1-2*x/n)**n));
finish;
%mend obsolete;

%macro runnulls;
%do i=1 %to 500;
    %QTLSIM(h0=0,mapscale=1,h=0.4,p0=0.15,r=0.728,seed=1235,rs=0.25,ssize=10000);
    proc append base=maxlod data=single;
    run;
    %QTLSIM(h0=0,mapscale=1,h=0.4,p0=0.05,r=0.9,seed=5431,rs=0.25,ssize=10000);
    proc append base=maxlod data=single;
    run;
%end;
%mend runnulls;

%macro analyse;
data _last_;
  set _last_;
  array qtl _numeric_;
  p1=0;p2=0;p3=0;p4=0;p5=0;
  do over qtl;
     if qtl<=1 then p1=p1+1;
     if qtl<=2 then p2=p2+1;
     if qtl<=3 then p3=p3+1;
     if qtl<=4 then p4=p4+1;
     if qtl<=5 then p5=p5+1;
  end;
run;
proc print;
run;
proc means n mean std;
run;
proc freq;
  table p1-p5;
run;
%mend;

%macro set1;
%do i=1 %to 1000;
    %QTLSIM(h0=1,mapscale=1,h=0.4,p0=0.15,r=0.7,seed=11111,rs=0.25,ssize=10000);
    proc append base=set1 data=single;
    run;
%end;
%analyse;
%mend set1;

%macro set2;
%do i=1 %to 1000;
    %QTLSIM(h0=1,mapscale=1,h=0.4,p0=0.05,r=0.9,seed=33333,rs=0.25,ssize=10000);
    proc append base=set2 data=single;
    run;
%end;
%analyse;
%mend set2;

title;
options nocenter nodate nonumber ps=500 ls=256;
* %runnulls;

filename logfile '/dev/null';
proc printto log=logfile;
run;
libname x '.';
%set1;
data x.set1;
  set set1;
run;
%set2;
data x.set2;
  set set2;
run;
